/*
 * Device driver for the via-pmu on Apple PowerBook 3400.
 *
 * The VIA (versatile interface adapter) interfaces to the PMU (power
 * management unit), a 6805 microprocessor core which controls the ADB
 * (Apple Desktop Bus) which connects to the keyboard and mouse.  The
 * PMU also controls system power and the RTC (real time clock) chip.
 *
 * This code emulates the cuda chip available on regular PowerMacs, and
 * it is called by the via-cuda driver
 *
 * Copyright (C) 1997 Fabio Riccardi.
 *
 * Copyright (C) 1996 Paul Mackerras.
 *
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991
 * Copyright 1996 1995 by Apple Computer, Inc. 1997 1996 1995 1994 1993 1992 1991
 */

#include <stdarg.h>
#include <linux/config.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <asm/prom.h>
#include <asm/cuda.h>
#include <asm/pmu.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/segment.h>

#ifdef CONFIG_XMON
#define DEBUG_PMU    
#endif

enum { FALSE, TRUE };

// globals shared with the cuda driver

extern volatile unsigned char *via;

/* VIA registers - spaced 0x200 bytes apart */
#define RS		0x200		/* skip between registers */
#define B		0		/* B-side data */
#define SR		(10*RS)		/* Shift register */
#define ACR		(11*RS)		/* Auxiliary control register */
#define IFR		(13*RS)		/* Interrupt flag register */
#define IER		(14*RS)		/* Interrupt enable register */


// added a state with respect to via-cuda

extern enum cuda_state {
    idle,
    sent_first_byte,
    sending,
    reading,
    read_done,
    awaiting_reply,
    asking_for_irq_data
} cuda_state;

extern struct cuda_request *current_req;

void cuda_input(unsigned char *buf, int nb, struct pt_regs *regs);


// to be called from via-cuda.c for polling
int is_pmu_interrupt(void) {
    return via[IFR] & ((1 << ifCB1) | (1 << ifSR));
}

// to be called from via-cuda.c after initialization
void enable_pmu_interrupts(void)
{
    via[IFR] = 0x7f; eieio();
    via[IER] = (1 << ifIRQ) | (1 << ifCB1) | (1 << ifSR); eieio();
}


// common wrap-up sequence for pmu->cuda requests

void pmu_start(void);

static inline void adb_done(struct cuda_request *req)
{
    if (req == NULL || req != current_req)
	panic("req != current_req");

    req->got_reply = 1;
    current_req = req->next;

    if (req->done)
	(*req->done)(req);

    cuda_state = idle;
    pmu_start();
}


// The following is somehow derived from Apple's MKLinux code,
// converts a cuda command into something edible for the PMU

enum {
    kPMUpMgrADB   	= 0x20,  	// send ADB command
    kPMUpMgrADBoff   	= 0x21,  	// turn ADB auto-poll off
    kPMUxPramWrite	= 0x32,  	// write extended PRAM byte(s)
    kPMUtimeRead  	= 0x38,  	// read the time from the clock chip
    kPMUxPramRead 	= 0x3A,  	// read extended PRAM byte(s)

    kPMUmaskInts	= 0x70,	 	//
    kPMUreadINT   	= 0x78,  	// get PMGR interrupt data
    kPMUPmgrPWRoff  	= 0x7E,  	// turn system power off
    kPMUresetCPU	= 0xD0,   	// reset the CPU

    PMU_readnvram	= 0x3b,
    PMU_writenvram	= 0x33,
};

static s8 StartSendCommand(u8 firstChar,
			   u8 *dataPointer1, u8 charCountS1,
			   u8 *dataPointer2, u8 charCountS2);

static u8	pmu_adb_flags	= 0;
static u16	pmu_device_map	= 0;	// bitmap of ADB devices

// offset for packing a reply message
enum { adb_offset = 1 };
int pmu_offset;
int adb_pseudo_offset;

static int pmu_handle_pseudo(struct cuda_request *req) 
{
    static u8 header[32];

    if (req->nbytes < 2) {
	printk(KERN_ERR "%s: bad length\n", __FUNCTION__);
	return -1;
    }

    req->reply[0] = ADB_RET_OK;		// status
    req->reply[1] = req->data[1];	// cuda command
    req->reply[2] = req->data[2];
    adb_pseudo_offset = 3;

    switch(req->data[1]) {
    case CUDA_AUTOPOLL:
	if (req->data[2] == 0) {
	    pmu_adb_flags = 0;
	    return StartSendCommand(kPMUpMgrADBoff, NULL, 0, NULL, 0);
	}
	header[0] = 0;
	header[1] = 0x86;
	header[2] = (u8) (pmu_device_map >> 8);
	header[3] = (u8) pmu_device_map;
	pmu_adb_flags = 0x02;
	return StartSendCommand(kPMUpMgrADB, header, 4, NULL, 0);

    case CUDA_SET_AUTO_RATE:
	adb_done(req);
	break;

    case CUDA_SET_DEVICE_LIST:
	pmu_device_map = (req->data[2] << 8) + req->data[3];
	adb_done(req);
	break;

    case CUDA_GET_TIME:
	return StartSendCommand(kPMUtimeRead, NULL, 0, NULL, 0);

    case CUDA_POWERDOWN:
	header[0] = 'M';
	header[1] = 'A';
	header[2] = 'T';
	header[3] = 'T';

	return StartSendCommand(kPMUPmgrPWRoff, header, 4, NULL, 0);

    case CUDA_RESET_SYSTEM:
	return StartSendCommand(kPMUresetCPU, NULL, 0, NULL, 0);
	break;

    default:
	printk(KERN_ERR "%s: unknown command\n", __FUNCTION__);
	return -1;
    }
    return 0;
}


// via-pmu serial line protocol

static inline void start_send(u8 theByte)
{
    via[ACR] |= 0x1C; eieio();		// set shift register to output
    via[SR] = theByte; eieio();		// give it the byte
    via[B] &= ~(1<<v2PMreq); eieio();	// assert /REQ
}

static inline void start_read(void)
{
    u8 theByte;
    
    // set shift register to input
    via[ACR] = (via[ACR] & ~0x10) | 0x0C; eieio();
    theByte = via[SR]; eieio();		// read a byte to reset shift reg
    via[B] &= ~(1<<v2PMreq); eieio();	// assert /REQ
}

static inline void do_req_lo(void)	{ via[B] |= (1<<v2PMreq); eieio(); }	// deassert /REQ line

static inline int is_ack_hi(void)	{ return via[B] & (1<<v2PMack); }
static inline int is_ack_lo(void)	{ return !(via[B] & (1<<v2PMack)); }
static inline int is_adb_irq(void)	{ return via[IFR] & (1<<ifCB1); }
static inline int is_sr_irq(void)	{ return via[IFR] & (1<<ifSR); }

static inline void ack_adb_irq(void)	{ via[IFR] = ((1<<ifIRQ) | (1<<ifCB1)); eieio(); }
static inline void ack_sr_irq(void)	{ via[IFR] = ((1<<ifIRQ) | (1<<ifSR)); eieio(); }

static inline u8 get_sr_byte(void)	{ u8 theByte = via[SR]; eieio(); return theByte; }


// called by via-cuda - processes cuda requests in the pmu fashion

static int pmu_pseudo;

/*
 * Called from via-cuda.c:cuda_send_request() with interrupts
 * disabled and cuda_state == idle.
 */
void pmu_start(void)
{
    struct cuda_request *req;
    static u8 header[32];

    /* get the packet to send */
    req = current_req;
    if (req == 0)
	return;
    
    switch(req->data[0]) {

    case PMU_PACKET:
	pmu_pseudo = TRUE;
	adb_pseudo_offset = 1;
	req->reply[0] = ADB_RET_OK;
	if (req->nbytes < 2
	    || StartSendCommand(req->data[1], &req->data[2],
				req->nbytes - 2, NULL, 0) < 0)
	    goto bad;
	break;

    case CUDA_PACKET:
	pmu_pseudo = TRUE;
	if (pmu_handle_pseudo(req) < 0)
	    goto bad;
	break;

    case ADB_PACKET:
	pmu_pseudo = FALSE;

	header[0] = req->data[1];
	header[1] = pmu_adb_flags;
	header[2] = req->nbytes - 2;

	StartSendCommand(kPMUpMgrADB, header, 3, &(req->data[2]), req->nbytes - 2);
	break;

    default:
	printk(KERN_ERR "%s: unknown request %d\n", __FUNCTION__,
	       req->data[0]);
    bad:
	req->reply[0] = ERROR_PACKET;
	req->reply[1] = req->data[0];
	req->reply[2] = req->data[1];
	req->reply_len = 3;
	adb_done(req);
	break;
    }
}

static int pmu_log = 0;

// for debugging

static char *cuda_state_names[] = {
    "idle",
    "sent_first_byte",
    "sending",
    "reading",
    "read_done",
    "awaiting_reply",
    "asking_for_irq_data"
};

static void pmu_complain(char *str, int x)
{
    printk(KERN_ERR "%s ERROR%d: cuda_state %s, IF %x, DB %x, ACR %x\n",
	   str, x, cuda_state_names[cuda_state],
	   via[IFR], via[B], via[ACR]);
#ifdef DEBUG_PMU
    xmon_printf("%s ERROR%d: cuda_state %s, IF %x, DB %x, ACR %x\n",
	   str, x, cuda_state_names[cuda_state],
	   via[IFR], via[B], via[ACR]);
    if (pmu_log > 0)
	xmon(0);
#endif
}


static u8	command[32];
static int	cmd_len = 0;
static int 	data_index;

extern s8 	cmdLengthTable[256];


static s8 StartSendCommand(u8 firstChar,
			      u8 *dataPointer1, u8 charCountS1,
			      u8 *dataPointer2, u8 charCountS2)
{
    int i;

    // validate pointers
    if ( ((charCountS1 != 0) && (dataPointer1 == NULL)) ||
	 ((charCountS2 != 0) && (dataPointer2 == NULL)) ) {
        printk(KERN_ERR "%s: Malformed message\n", __FUNCTION__);
	return -1;
    }

    if ( (firstChar != 0xE1) &&				// validate length
	 (cmdLengthTable[firstChar] != -1) &&
	 (cmdLengthTable[firstChar] != (charCountS1 + charCountS2)) ) {
        printk(KERN_ERR "%s: Malformed message length: %d(0x%x) %d %d %d\n",
	       __FUNCTION__, firstChar, firstChar,
	       cmdLengthTable[firstChar], charCountS1, charCountS2);
	return -1;
    }

    cmd_len = 0;
    command[cmd_len++] = firstChar;

    if ( cmdLengthTable[firstChar] < 0 )		// should we send a length byte?
	command[cmd_len++] = (u8) (charCountS1 + charCountS2);

    for ( i = 0; i < charCountS1; i++ )			// send data bytes
	command[cmd_len++] = *dataPointer1++;

    for ( i = 0; i < charCountS2; i++ )			// send more data bytes
	command[cmd_len++] = *dataPointer2++;

    data_index = 1;
    cuda_state = sending;

    start_send(command[0]);

    return 0;
}


// cuda-style interrupt handler
//
// mostly interrupt driven, except for waiting for ack_hi, which uses
// busy waiting, since it seems that there is no interrupt available
// for ack_hi

// don't complain during initialization
static int pmu_initted = FALSE;

static void startTimer(Timer * timer, u32 duration);
static int PMUtimeout(Timer * timer);

static unsigned char cuda_rbuf[16];
static unsigned char *reply_ptr = NULL;
static int bytes_to_read = 0;
static int reading_reply = FALSE;
static int bytes_read = 0;
static int adb_int_pending;

extern s8 	rspLengthTable[256];
static struct cuda_request set_bright_cmd = {
    { PMU_PACKET, 0x41, }, 3,
};

void via_pmu_interrupt(int irq, void *arg, struct pt_regs *regs)
{
    struct cuda_request *req;
    u8 theByte;
    int retries;

#ifdef DEBUG_PMU
    if (pmu_log >= 2) {
	xmon_printf("--- %s: cuda_state %s, btr %d, br %d, IF %x, DB %x, ACR %x\n",
		    __FUNCTION__, cuda_state_names[cuda_state], bytes_to_read, bytes_read,
		    via[IFR], via[B], via[ACR]);
	xmon_printf(" reading_reply=%d pmu_offset=%d data_index=%d cmd_len=%d\n",
		    reading_reply, pmu_offset, data_index, cmd_len);
    }
#endif

    if (is_adb_irq()) {
	ack_adb_irq();
	adb_int_pending = 1;
#ifdef DEBUG_PMU
	if (pmu_log == 1)
	    xmon_printf("pmu: cb1 interrupt\n");
#endif
    }
    if (is_sr_irq()) {
	ack_sr_irq();
	/* If reading, grab the byte from the shift register now */
	if ((via[ACR] & 0x10) == 0)
	    theByte = get_sr_byte();

	if (!is_ack_lo()) {
	    pmu_complain(__FUNCTION__ ": out of step", 2);
	    return;
	}

	for (retries = 0; retries++ < 5 && is_ack_lo(); ) {
	    Timer timer;

	    do_req_lo();

	    // wait up to 32 milliseconds for Ack signal from PG&E to go high
	    startTimer(&timer, TIMEOUT32MS);
	    while (is_ack_lo() && !PMUtimeout(&timer))
		;
	}
	if (is_ack_lo()) {
	    pmu_complain(__FUNCTION__ ": transmission error", 1);
	    return;
	}

	switch (cuda_state) {
	case asking_for_irq_data:
#ifdef DEBUG_PMU
	    if (pmu_log)
		xmon_printf("asking_for_irq_data: reading_reply=%d\n",
			    reading_reply);
#endif
	    reply_ptr = reading_reply ? current_req->reply : cuda_rbuf;
	    pmu_offset = adb_offset;
	    bytes_to_read = -1;
	    bytes_read = 0;
	    cuda_state = reading;
	    pmu_pseudo = 0;
	    start_read();
	    break;

	case sent_first_byte:
#if 0
	    if (is_adb_irq()) {
		// collision, abort sending... this never seems to happen anyway...
		via[ACR] |= 0x0C; eieio();	// set shift register to input
		theByte = via[SR]; eieio();	// read a byte to reset shift reg
		via[ACR] &= ~0x10; eieio();

		pmu_complain(__FUNCTION__ ": *** COLLISION ***", 3);

		cuda_state = idle;
		break;
	    } else
#endif
	    {
		cuda_state = sending;
	    }
	    // fall through

	case sending:
	    req = current_req;
	    if (data_index < cmd_len) {
		start_send(command[data_index++]);
		break;
	    }
#ifdef DEBUG_PMU
	    if (pmu_log) {
		int i;
		xmon_printf("pmu: sent cmd");
		for (i = 0; i < data_index; ++i)
		    xmon_printf(" %.2x", command[i]);
	    }
#endif
	    req->sent = 1;
	    if (pmu_pseudo) {
		bytes_to_read = rspLengthTable[command[0]];
		    
		if (bytes_to_read != 0) {
		    if (bytes_to_read > 1)
			bytes_to_read--;
#ifdef DEBUG_PMU
		    if (pmu_log)
			xmon_printf(" expecting %d bytes", bytes_to_read);
#endif

		    reply_ptr = current_req->reply;
		    pmu_offset = adb_pseudo_offset;

		    reading_reply = TRUE;
		    bytes_read = 0;
		    cuda_state = reading;
		    start_read();
		} else {
		    req->reply_len = 0;
		    adb_done(req);
		}
	    } else if (req->reply_expected) {
		cuda_state = awaiting_reply;
#ifdef DEBUG_PMU
		if (pmu_log)
		    xmon_printf(" awaiting reply");
#endif
	    } else
		adb_done(req);
#ifdef DEBUG_PMU
	    if (pmu_log)
		xmon_printf("\n");
#endif
	    break;

	case reading:
	    if (bytes_to_read == -1) {
		bytes_to_read = (s8) theByte;

		if (bytes_to_read < 0 || bytes_to_read > 16) {
		    printk(KERN_ERR "%s: bad length %d\n",
			   __FUNCTION__, bytes_to_read);
		    
		    bytes_to_read = 0;
		}
	    } else
		reply_ptr[pmu_offset + bytes_read++] = theByte;

	    if (bytes_to_read > bytes_read) {
		start_read();
		break;
	    }

#ifdef DEBUG_PMU
	    if (pmu_log) {
		int i;
		xmon_printf("read %d bytes:", bytes_read);
		for (i = 0; i < bytes_read; ++i)
		    xmon_printf(" %.2x", reply_ptr[pmu_offset+i]);
		xmon_printf("  (reply_ptr=%x pmu_offset=%x rd_rply=%d)\n",
			    reply_ptr, pmu_offset, reading_reply);
	    }
#endif
	    if (pmu_pseudo) {
		req = current_req;
		req->reply_len = pmu_offset + bytes_read;
		adb_done(req);
		break;
	    }

	    if (reply_ptr[pmu_offset] & PMU_ADB_INT_MASK) {
		if (bytes_read < 1)
		    panic("ADB Interrupt: bad length\n");
		    
		if (reading_reply) {
		    req = current_req;
		    req->reply_len = bytes_read + pmu_offset;

		    if (reply_ptr[pmu_offset] & 0x04)
			printk(KERN_ERR "unsolicited interrupt while waiting for reply!\n");
		    if (bytes_read > 2)
			req->reply[0] = ADB_RET_OK;
		    else
			req->reply[0] = ADB_RET_TIMEOUT;

		    bytes_to_read = bytes_read = 0;
		    reading_reply = 0;
		    adb_done(req);
		} else {
		    if (!(reply_ptr[pmu_offset] & 0x04))
			printk(KERN_ERR "unexpected adb reply interrupt!\n");

		    cuda_rbuf[0] = ADB_PACKET;
		    if (pmu_initted)
			cuda_input(cuda_rbuf, bytes_read + pmu_offset, regs);
		    cuda_state = idle;
		    bytes_to_read = bytes_read = 0;
		    pmu_start();
		}
	    } else {
		static int show_pmu_ints;
		if (show_pmu_ints
		    && !(reply_ptr[pmu_offset] == 0x80 && bytes_read == 1)) {
		    int i;
		    printk("pmu intr");
		    for (i = 0; i < bytes_read; ++i)
			printk(" %x", reply_ptr[pmu_offset+i]);
		    printk("\n");
		}

		if (reply_ptr[pmu_offset] == 0x08) {
		    /* sound/brightness buttons pressed */
		    if (set_bright_cmd.got_reply) {
			if (reply_ptr[pmu_offset+1] < 4)
			    set_bright_cmd.data[2] = 0x7f;
			else
			    set_bright_cmd.data[2] = 0x80 -
				(reply_ptr[pmu_offset+1] >> 1);
			cuda_send_request(&set_bright_cmd);
		    }
		}
		bytes_to_read = bytes_read = 0;
		reading_reply = 0;
		if (reading_reply) {
		    cuda_state = awaiting_reply;
		} else {
		    cuda_state = idle;
		    pmu_start();
		}
	    }
	    break;

	default:
	    // bad state
	    pmu_complain(__FUNCTION__ ": bad state on sr_irq", 4);
	    break;
	}
    }
    if (adb_int_pending
	&& (cuda_state == idle || cuda_state == awaiting_reply)) {
	adb_int_pending = 0;
	reading_reply = cuda_state == awaiting_reply;
	cuda_state = asking_for_irq_data;
	start_send(kPMUreadINT);
    }
}


int init_via_pmu(void)
{
    u8 mask_buf[1];
    struct cuda_request init_req;

    //    mask_buf[0] = PMU_ADB_INT_MASK;		// mask out all ints except ADB
    mask_buf[0] = 0xff;

    current_req = &init_req;
    current_req->nbytes = 1;
    current_req->done = NULL;
    current_req->reply_expected = 0;
    current_req->next = 0;
    current_req->sent = 0;
    current_req->got_reply = 0;
    current_req->reply_len = 0;
    
    StartSendCommand(kPMUmaskInts, mask_buf, 1, 0, 0);
    while(cuda_state != idle)
	via_pmu_interrupt(0, 0, 0);
    
    current_req = NULL;
    
    // sync with PMU and ack all pending irqs
    do {
	if (is_adb_irq())
	    ack_adb_irq();
	cuda_state = asking_for_irq_data;
	start_send(kPMUreadINT);
	while (cuda_state != idle)
	    via_pmu_interrupt(0, 0, 0);
    } while (cuda_rbuf[adb_offset]);

    set_bright_cmd.got_reply = 1;
    set_bright_cmd.data[0] = PMU_PACKET;
    set_bright_cmd.data[1] = 0x41;
    set_bright_cmd.nbytes = 3;
    set_bright_cmd.done = 0;

    pmu_initted = TRUE;

    return 1;
}


// the following code is what remains of the original Apple's hack

// *************************************************************************************
// Start a timer by copying the current value of the timebase
// *************************************************************************************

inline static unsigned int mftb(void)
{
    unsigned int result;
    __asm__ volatile("mftb %0" : "=r" (result));
    return result;
}

inline static unsigned int mftbu(void)
{
    unsigned int result;
    __asm__ volatile("mftbu %0" : "=r" (result));
    return result;
}

static void startTimer(Timer * timer, u32 duration)
{
    u32 temp;

    while ( 1 )
    {
	timer->upper = mftbu();
	timer->lower = mftb();
	temp = mftbu();
	if ( temp == timer->upper ) break;
    }
    if ( (timer->lower += duration) < timer->lower ) ++timer->upper;
    timer->lower += duration;
}


// *************************************************************************************
// Check to see if a timer has expired by comparing the sum of its start value
// and its duration to the current value of the timebase
// *************************************************************************************

static int PMUtimeout(Timer * timer)
{
    u32 currentUpper;
    u32 currentLower;
    u32 temp;

    while ( 1 ) {
	currentUpper = mftbu();
	currentLower = mftb();
	temp = mftbu();
	if ( temp == currentUpper ) break;
    }

    if ( currentLower > timer->lower )
	return currentUpper < timer->upper ? FALSE : TRUE;

    return currentUpper > timer->upper ? TRUE : FALSE;
}


//  This tables are used to determine how to handle the command send and reply:

// =0:	no reply should be expected.
// =1:  only a reply byte will be sent (this is a special case for a couple of commands)
// <0:	a reply is expected and the PMGR will send a count byte.
// >1:	a reply is expected and the PMGR will not send a count byte,
//	but the count will be (value-1).

s8 cmdLengthTable[256] = {
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,	// 0x00 - 0x0F
     1, 1,-1,-1,-1,-1,-1,-1, 0, 0,-1,-1,-1,-1,-1, 0,	// 0x10 - 0x1F
    -1, 0, 2, 1, 1,-1,-1,-1, 0,-1,-1,-1,-1,-1,-1,-1,	// 0x20 - 0x2F
     4,20, 2, 3,-1,-1,-1,-1, 0, 0, 1, 2,-1,-1,-1,-1,	// 0x30 - 0x3F
     1, 1,-1,-1,-1,-1,-1,-1, 0, 0,-1,-1, 1,-1,-1,-1,	// 0x40 - 0x4F
     1, 0, 2, 2,-1, 1, 3, 1, 0, 1, 0, 0, 0,-1,-1,-1,	// 0x50 - 0x5F
     2,-1, 2, 0,-1,-1,-1,-1, 0, 0, 0, 0, 0, 0,-1,-1,	// 0x60 - 0x6F
     1, 1, 1,-1,-1,-1,-1,-1, 0, 0,-1,-1,-1,-1, 4, 4,	// 0x70 - 0x7F
     4,-1, 0,-1,-1,-1,-1,-1, 0,-1,-1,-1,-1,-1,-1,-1,	// 0x80 - 0x8F
     1, 2,-1,-1,-1,-1,-1,-1, 0, 0,-1,-1,-1,-1,-1,-1,	// 0x90 - 0x9F
     2, 2, 2, 4,-1, 0,-1,-1, 1, 1, 3, 2,-1,-1,-1,-1,	// 0xA0 - 0xAF
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,	// 0xB0 - 0xBF
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,	// 0xC0 - 0xCF
     0,-1,-1,-1,-1,-1,-1,-1, 1, 1,-1,-1, 0, 0,-1,-1,	// 0xD0 - 0xDF
    -1, 4, 0,-1,-1,-1,-1,-1, 3,-1, 0,-1, 0,-1,-1, 0,	// 0xE0 - 0xEF
    -1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1	// 0xF0 - 0xFF
};

s8 rspLengthTable[256] = {
    0, 0, 0, 0, 0, 0, 0, 0,-1,-1,-1,-1,-1,-1,-1,-1,	// 0x00 - 0x0F
    0, 0, 0, 0, 0, 0, 0, 0, 2, 2,-1,-1,-1,-1,-1, 0,	// 0x10 - 0x1F
    0, 0, 0, 0, 0, 0, 0, 0,-1,-1,-1,-1,-1,-1,-1,-1,	// 0x20 - 0x2F
    0, 0, 0, 0, 0, 0, 0, 0, 5,21, 1, 1,-1,-1,-1,-1,	// 0x30 - 0x3F
    0, 0, 0, 0, 0, 0, 0, 0, 2, 2,-1,-1, 0,-1,-1,-1,	// 0x40 - 0x4F
    0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 3, 3,-1,-1,-1,-1,	// 0x50 - 0x5F
    0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 3, 9,-1,-1,-1,-1,	// 0x60 - 0x6F
    0, 0, 0, 0, 0, 0, 0, 0,-1,-1,-1,-1,-1,-1, 1, 1,	// 0x70 - 0x7F
    0, 0, 0, 0, 0, 0, 0, 0, 6,-1,-1,-1,-1,-1,-1,-1,	// 0x80 - 0x8F
    0, 0, 0, 0, 0, 0, 0, 0, 2, 2,-1,-1,-1,-1,-1,-1,	// 0x90 - 0x9F
    0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0,-1,-1,-1,-1,	// 0xA0 - 0xAF
    0, 0, 0, 0, 0, 0, 0, 0,-1,-1,-1,-1,-1,-1,-1,-1,	// 0xB0 - 0xBF
    0, 0, 0, 0, 0, 0, 0, 0,-1,-1,-1,-1,-1,-1,-1,-1,	// 0xC0 - 0xCF
    0, 0, 0, 0, 0, 0, 0, 0, 2, 2,-1,-1, 2,-1,-1,-1,	// 0xD0 - 0xDF
    0, 0, 1, 0, 0, 0, 0, 0,-1,-1, 2,-1,-1,-1,-1, 0,	// 0xE0 - 0xEF
    0, 0, 0, 0, 0, 0, 0, 0,-1,-1,-1,-1,-1,-1,-1,-1	// 0xF0 - 0xFF
};
