// Copyright (c) 2001-2002 TiVo Inc.
//
// Author: Andy Goodman 

#ifndef __QDMA_H__
#define __QDMA_H__
/*
 * Includes
 */
#include "linkq.h"


/*
 * Defines
 */


/*
 * Frags should be marked with their transfer state.
 * FIXME...not yet implemented.
 */
typedef struct _DmaChReqT DmaChReqT;

typedef enum {
  tyNotTransferred = 0,
  tyTransferring = 1,
  tyTransferred = 3,
} DmaChFragStateT;

/*
 * The length must be less than 2**30-1 to leave 2 bits for
 * the DmaChFragStateT in the msb's.
 *
 * The buffers (buf) MUST BE ALLIGNED ON 4 BYTE BOUNDARIES.
 * The lengths (len) are the number of BYTES in the fragment
 */
typedef struct _DmaChFragT {
  unsigned char *buf;
  unsigned long len; 
}DmaChFragT;


/*
 * This is written as a state machine to
 * provide useful debug info.
 */
typedef enum {
  tyNextIrq = 1,
  tyIdleIrq = 2,
  tyNextAndIdleIrq = 3,
  tyStartMove,
  tyReady,
  tyStop,
  tyEnqueue,
  tyRingStart,
  tyRingStop
}DmaChEventT;
 





/*
 * This is the request structure that is
 * passed to DmaChEnqueue().
 *
 * Only fill in the frags and numFrags fields.
 * The other fields are used by the engine.
 */
struct _DmaChReqT {
  DmaChReqT *next;
  DmaChFragT *fragp; /* The actual frag table. */
  int numFrags;  /* Number of frags in this DMA request. */
  int finished;
  void *refcon;	
  int maxFrags;  /* Maximum allowable number of frags in a DMA request. */
};


/*
 * These are the DMA channels
 */
typedef enum {
  tyDmaChStart=-1, /* keep at beginning */

  /* MediaSwitch 1 */
  tyDmaChIde,
  tyDmaChVideoIn,
  tyDmaChAudioIn,
  tyDmaChVbi,
  tyDmaChEventsIn,
  tyDmaChVideoOut,
  tyDmaChAudioOut,
  tyDmaChOsd,
  /* MediaSwitch 2 */
  tyDmaChVideoIn2,
  tyDmaChDigVideoIn1 = tyDmaChVideoIn2, /* MS 4 */
  tyDmaChAudioIn2,
  tyDmaChDigAudioIn1 = tyDmaChAudioIn2, /* MS 4 */
  tyDmaChVbi2,
  tyDmaChEventsIn2,
  tyDmaChOSLinkIn,
  tyDmaChOSLinkOut,
  tyDmaChAudioOut2,
  tyDmaChOsd2,
  tyDmaChILinkOut,
  /* MediaSwitch 4 */
  tyDmaChDigVideoIn2,
  tyDmaChDigAudioIn2,
  tyDmaChOSLinkIn2,
  tyDmaChOSLinkOut2,
  tyDmaChIdeSG,

  tyDmaChEnd /* keep at end */
}TyDmaChT;


typedef struct _DmaChParmsT {
  void (*Notify)(DmaChReqT *req, void *pNotify);
  void *pNotify;
} DmaChParmsT;


/*
 * This is the structure passed to DmaChInit function.
 */
typedef struct _DmaChInitT {
  char *name;          /*
			* Dma Channel name
			*/
  bool fromListToPort; /* TRUE if transferring from List to Port, false
			* otherwise
			*/
  bool onMediaSwitch2; /* TRUE if on MediaSwitch2, False if on MediaSwitch1 */
  U32 curAddrReg;
  U32 curCountReg;
  U32 nextAddrReg;
  U32 nextCountReg;
  U32 nextEmptyMask;
  U32 idleMask;
  U32 resetFifoMask;
}DmaChInitT;


/*
 * Globals
 */

#ifdef ALLOC_ILINK_DMA_BUFFERS
/*
  This is a range of continuous physical pages managed by the i.LINK
  driver. This was the simplest way to get CONSECUTIVE PHYSICAL pages
  in order to make large DMA fragments. Alternative methods of
  __get_dma_pages() and kmalloc(), with GPF_DMA, do not return
  consecutive physical pages!

  This variable is configured only if ALLOC_ILINK_DMA_BUFFERS is
  defined.
 */
extern unsigned ilink_dma_size;
extern char ilink_dma_buffers[];
#endif

/*
 * Functions
 */

/*
 * Has a request finished?
 *
 * Once this is called, the request will be reset to a
 * non finished state
 */
static inline bool DmaChReqIsFinished(DmaChReqT *me)
{
  if  (me->finished ){
    me->finished = 0;
    return TRUE;
  }
  return FALSE;
}



/*
 * Called at system initialization
 */
extern void DmaChInitAll(void);

/*
 * Normally DmaChInitAll calls DmaChInit for each of the DMA channels. 
 * DmaChInit is here exported in case it needs to be reinitialized.
 */
void DmaChInit(TyDmaChT channel, DmaChInitT *init);


/*
 * You can always enqueue something nueue.
 *
 * The buffers in the requests (buf) MUST BE ALLIGNED ON 4 BYTE BOUNDARIES.
 * The lengths in the requests(len) are the number of BYTES in the fragment 
 */
extern void DmaChEnqueue(TyDmaChT channel,DmaChReqT *req);
extern void DmaListChEnqueue(TyDmaChT listchannel, TyDmaChT channel, DmaChReqT *req);


/*
 * Starts/stops  transfer. Returns -1 if
 * if there are no requests enqueued.
 */
extern int DmaChStart(TyDmaChT channel);



/*
 * Set channel parameters if you need to...
 *
 * Returns -1 if channel not idle or stopped.
 */
extern int DmaChSetParameters(TyDmaChT channel, DmaChParmsT *parms);


/*
 * Returns -1 if channel is idle or stopped.
 */
extern int DmaChStop(TyDmaChT channel,bool immediate);

/*
 * Dump a request
 */
extern void DmaChReqPrint(DmaChReqT *me);




/*
 * Called from either of the IRQ's. Triggers next transfer
 * if there is one. Check state afterwords to findout
 * if list is finished.
 *
 * These are called by the kernel or by a test program on the raw
 * hardware.
 */
extern int DmaChNextIrq(TyDmaChT channel);
extern int DmaChIdleIrq(TyDmaChT channel);
extern int DmaChNextAndIdleIrq(TyDmaChT channel);

extern U32 DmaChGetIrqs(TyDmaChT channel);
/*
 * Returns TRUE if the channel has finished all requests and FALSE if busy
 */
extern bool DmaChIdle(TyDmaChT channel);



/*
 * Functions for ringing DMA's. You can only switch from buffered 
 * to ringing when the channel is idle.
 *
 * If a  notify function is registered, it will be called on each ring of
 * the buffer.
 *
 * DmaChGetAddr() gets the current DMA address out of the mswitch
 * register. It only works for ringing DMA's currently.
 */
extern int DmaChGetAddr(TyDmaChT channel, U32 *addr);
extern int DmaChRingStart(TyDmaChT channel, U32 buf, U32 length);
extern int DmaChRingStop(TyDmaChT channel);
extern void DmaChDebugPrint(TyDmaChT channel, U32 on);



#define QDMA_MAJOR 91 /* major number */


/*
 * IOCTLS
 */

/*
 * Debug levels
 */
#define DEBUG_STATES 1
#define DEBUG_FRAGS  2
#define DEBUG_IDLE   4
#define DEBUG_ALL    (1<<31)

/*
 * IOCTL -- set debug level
 */
#define DMA_CH_DEBUG_LEVEL 0
#define DMA_CH_DUMP_STATE  1

typedef struct _DmaChDebugT {
  TyDmaChT channel;
  int level;
} DmaChDebugT;
extern void DmaChDumpState(int channel);
#endif
